"""All classes that implement transitions."""
# coding=utf-8

import abc
import typing
from functools import total_ordering

from .exceptions import ConfigurationError, StateTransitionError


if typing.TYPE_CHECKING:  # pragma: no cover
  from . import engine  # pylint: disable=unused-import

__all__ = [
  'TransitionOption',
  'SimpleTransition',
  'Transition',
  'StateType',
  'ContextType',
  'DescribedTransition',
  'SimpleDescribedTransition',
]

StateType = typing.TypeVar('StateType')
ContextType = typing.TypeVar('ContextType')

TransitionOption = typing.NamedTuple(
  'TransitionOption',
  (
    ("transition", "Transition"),
    ("transition_name", str),
    ("from_state", StateType),
    ("to_state", StateType),
    ("allowed", bool),
    # If transition is disallowed this should tell you why
    ("reason", typing.Optional[str]),
  )
)


class Transition(typing.Generic[StateType, ContextType]):
  """Transition base class."""

  def __init__(  # pylint: disable=unused-argument
      self,
      transition_name: str,
      from_state: typing.Optional[StateType],
      to_state: typing.Optional[StateType],
      **kwargs
  ):
    """

    :param transition_name: Name of transition, this name must be unique.
    :param from_state: Initial state. None means "from any state".
    :param to_state: Final state.
    :param kwargs: This class ignores any extra kwargs.
    """
    self.from_state = from_state
    self.to_state = to_state
    self.transition_name = transition_name

  @abc.abstractmethod
  def can_transition(
      self,
      machine: 'engine.MachineConfiguration',
      from_state: typing.Optional[StateType],
      context: ContextType = None,
      extra_args: dict = None
  ) -> TransitionOption:
    """Return object that describes whether this transition is currently legal. """
    raise NotImplementedError

  def transition(
      self,
      machine: 'engine.MachineConfiguration',
      from_state: typing.Optional[StateType],
      context: ContextType = None,
      extra_args: dict = None
  ) -> StateType:
    """Perform the transition. """
    option = self.can_transition(
      machine=machine,
      from_state=from_state,
      context=context,
      extra_args=extra_args
    )
    if not option.allowed:
      raise StateTransitionError(
        "Illegal transition '{}'".format(option.reason)
      )
    return option.to_state

  def __eq__(self, other):
    return self.transition_name == other.transition_name

  def __hash__(self):
    return hash(self.transition_name)


@total_ordering
class DescribedTransition(Transition[StateType, ContextType]):
  """
  Transition with label, description and priority.

  Priority is used to sort transitions
  """

  def __init__(
      self,
      transition_name: str,
      from_state: typing.Optional[StateType],
      to_state: typing.Optional[StateType],
      label: str,
      priority: int = None,
      description: str = None,
      **kwargs
  ):
    super().__init__(transition_name, from_state, to_state, **kwargs)
    self.label = label
    self.description = description
    if priority is None:
      priority = to_state
    self.priority = priority

  def __ge__(self, other):
    return (
      (self.priority, self.label, self.transition_name) >
      (other.priority, other.label, other.transition_name)
    )


class SimpleTransition(Transition[StateType, ContextType]):
  """Transition that can be performed always."""

  def __init__(self, transition_name: str, from_state: typing.Optional[StateType],
               to_state: typing.Optional[StateType], **kwargs):
    super().__init__(transition_name, from_state, to_state, **kwargs)
    if to_state is None:
      raise ConfigurationError("Need to set to_state for simple transition")

  def _on_successful_transition(
      self,
      machine: 'engine.MachineConfiguration',
      from_state: typing.Optional[StateType],
      to_state: typing.Optional[StateType],
      context: ContextType = None,
      extra_args: dict = None
  ):
    """
    Callback called from ``transition`` on a valid transition.

    Called after ``can_transition``` allowed transition.
    """

  def _make_transition_option(  # pylint: disable=unused-argument
      self,
      allowed: bool,
      **kwargs
  ):
    """Create transition option object, setting sane defaults."""

    def get(name):
      """Read field from kwargs and if None get from self."""
      value = kwargs.get(name, getattr(self, name, None))
      if value is None:
        raise ValueError("Transition property {} is undefined".format(name))
      return value

    if not (allowed or kwargs.get('reason')):
      raise ValueError("Please specify reason if transition is not allowed.")

    return TransitionOption(
      transition_name=self.transition_name,
      from_state=get('from_state'),
      to_state=get('to_state'),
      allowed=allowed,
      reason=kwargs.get('reason'),
      transition=self
    )

  def can_transition(  # pylint: disable=unused-argument
      self,
      machine: 'engine.MachineConfiguration',
      from_state: typing.Optional[StateType],
      context: ContextType = None,
      extra_args: dict = None
  ) -> TransitionOption:
    """Always allow transition."""

    if extra_args is None:
      extra_args = {}

    return self._make_transition_option(
      allowed=True,
      from_state=from_state,
      context=context,
      **extra_args
    )

  def transition(
      self,
      machine: 'engine.MachineConfiguration',
      from_state: typing.Optional[StateType],
      context: ContextType = None,
      extra_args: dict = None
  ) -> StateType:
    """Perform the transition. """
    to_state = super().transition(
      machine=machine,
      from_state=from_state,
      context=context,
      extra_args=extra_args
    )
    self._on_successful_transition(
      to_state=to_state,
      machine=machine,
      from_state=from_state,
      context=context,
      extra_args=extra_args
    )
    return to_state


class SimpleDescribedTransition(
    DescribedTransition[StateType, ContextType],
    SimpleTransition[StateType, ContextType]
):
  """Transition that can be performed always and it has a description."""
  pass



"""
All exceptions raised by yapfsm
"""

# coding=utf-8

__all__ = [
  "FSMError", "ConfigurationError", "StateTransitionError",
]


class FSMError(Exception):
  """Base exception."""
  def __init__(self, message):
    super().__init__()
    self.message = message


class ConfigurationError(FSMError):
  """Raised when FSM has configuration issues (e.g. duplicate transition)."""
  pass


class StateTransitionError(FSMError):
  """Raised when user requested illegal transition."""
  pass



"""Actual FSM implementation."""

# coding=utf-8

import typing

from collections import defaultdict

from .exceptions import ConfigurationError, StateTransitionError

from . import transition

__all__ = [
  'TransitionType',
  'MachineConfiguration',
  'Machine'
]


TransitionType = typing.TypeVar('TransitionType')


def first_true(iterable, default=False, pred=None):
  """Returns the first true value in the iterable.

  If no true value is found, returns *default*

  If *pred* is not None, returns the first item
  for which pred(item) is true.

  """
  # first_true([a,b,c], x) --> a or b or c or x
  # first_true([a,b], x, f) --> a if f(a) else b if f(b) else x
  return next(filter(pred, iterable), default)


class MachineConfiguration(typing.Generic[transition.StateType, TransitionType]):
  """
  Immutable FSM configuration, can be shared between many state machines.
  Stores all possible transitions.
  """
  def __init__(self):
    super().__init__()
    self.transitions = defaultdict(list)
    """Maps state to list of legal transitions."""
    self.flat_transitions = set()
    """All transitions in set. Used to detect duplicate transitions."""

  def add_transition(self, trans: TransitionType):
    """
    Adds transition to this configuration.

    :raises ConfigurationError: When such transition exists.
    """
    if trans in self.flat_transitions:
      raise ConfigurationError("Transition already present")
    self.transitions[trans.from_state].append(trans)
    self.flat_transitions.add(trans)

  def get_transitions_from_state(
      self,
      from_state: transition.StateType
  ) -> typing.Iterable[transition.Transition]:
    """
    Returns all possible transitions from state.
    :param from_state: State for which to find transitions. \
    """
    global_transitions = self.transitions.get(None, list())
    local_transitions = self.transitions.get(from_state, list())
    return sorted(global_transitions + local_transitions)

  def transition_by_name(
      self,
      transition_name: str,
      from_state: transition.StateType
  ) -> TransitionType:
    """
    Gets transition by name.
    """
    def match_by_name(trans: TransitionType):
      """Predicate to find transition"""
      return trans.transition_name == transition_name
    result = first_true(
      self.get_transitions_from_state(from_state), None, match_by_name
    )
    if result is None:
      raise ConfigurationError("No such transition")
    return result


class Machine(typing.Generic[transition.StateType, transition.ContextType]):
  """Mutable FSM machine."""

  def __init__(
      self,
      initial_state: transition.StateType,
      machine: MachineConfiguration,
      context: transition.ContextType
  ):
    self.machine = machine
    self.context = context
    self.current_state = initial_state

  def update_state(self, new_state):
    """
    Updates state of the machine. If you need to do some custom action on state update
    do it here.
    """
    self.current_state = new_state

  def perform_transition(self, transition_name: str, extra_args: dict = None):
    """Performs state transition from current state."""
    try:
      trans = self.machine.transition_by_name(transition_name, self.current_state)
    except ConfigurationError as exc:
      raise StateTransitionError(exc.message) from exc
    next_state = trans.transition(
      machine=self.machine,
      from_state=self.current_state,
      context=self.context,
      extra_args=extra_args
    )
    self.update_state(next_state)

  def can_transition(
      self,
      transition_name: str,
      extra_args: dict = None
  ) -> transition.TransitionOption:
    """
    Checks if transition can be performed.

    :param transition_name: Name of the transition
    :param extra_args: Any extra arguments
    """
    try:
      trans = self.machine.transition_by_name(transition_name, self.current_state)
    except ConfigurationError:
      return transition.TransitionOption(
        transition=None,
        transition_name=transition_name,
        from_state=self.current_state,
        to_state=None,
        allowed=False,
        reason="No such transition"
      )
    return trans.can_transition(
      machine=self,
      from_state=self.current_state,
      context=self.context,
      extra_args=extra_args
    )

  def get_possible_transitions(
      self,
      extra_args: dict = None
  ) -> typing.Iterable[transition.TransitionOption]:
    """
    Returns all transitions for current state, along with information if transition
    is currently legal.
    """
    return [
      t.can_transition(
        machine=self.machine,
        from_state=self.current_state,
        context=self.context,
        extra_args=extra_args
      )
      for t in self.machine.get_transitions_from_state(self.current_state)
    ]
